﻿//-----------------------------------------------------------------------------------------------
// SatPC32 sends data over DDE: 
//	DDE service: SatPC32
//	DDE topic: SatPcDdeConv
//	DDE item: SatPcDdeItem
//
// and sends one of the following messages (approx every second):
//		No satellite selected, or the one selected has elevation below 0 (notice the extra \n !)
//			** NO SATELLITE **\n\r\n
//
//		For QO-100 below there is no CAT info in the doppler.sqf file for SatPC32, so both TX and RX frequencies are 0
//			SNQO-100 AZ160,4 EL24,9 UP0 UMLSB DN0 DMUSB MA254,3 RR-0,0003316\r\n
//
//		AO-73 with CAT data in doppler.sqf file:
//			SNAO-73 AZ91,7 EL1,9 UP435132865 UMLSB DN145962393 DMUSB MA209,5 RR-4,9154719\r\n
//			SNAO-73 AZ96,0 EL0,3 UP435132355 UMLSB DN145962564 DMUSB MA207,7 RR-5,2667847\r\n
//			SNAO-73 AZ85,6 EL4,0 UP435133668 UMLSB DN145962124 DMUSB MA211,6 RR-4,3627629\r\n
//-----------------------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using Microsoft.VisualBasic;
using NDde.Client; 
using SuperSimpleTcp;
using SimpleUdp;

namespace ACLogSatPC32 {
	public partial class Form1 : Form {

		private DdeClient clientSatPC32;	// DDE connection to SatPC32
		public SimpleTcpClient clientACLog; // Socket connection to ACLog (for its client API)
		public UdpEndpoint clientUDP;       // UDP client for sending Az/El to the OZ9AAR URC Rotorcontroller
		private bool initDone = false;

		// Define names for the FREQ_RX and BAND_RX fields in ACLog
		private	String FREQ_RX = "OTHER7";
		private String BAND_RX = "OTHER8";

		public Form1() {
			InitializeComponent();

			// Create client for SatPC32 DDE connection
			clientSatPC32 = new DdeClient("SatPC32", "SatPcDdeConv", this);
			clientSatPC32.Advise += clientSatPC32_Advise;
			clientSatPC32.Disconnected += clientSatPC32_Disconnected;

			// Create socket client for ACLog connection

			clientACLog = new SimpleTcpClient("127.0.0.1", Convert.ToInt32(Settings1.Default.ACLogPort));
			clientACLog.Events.Connected	+= clientACLog_Connected;
			clientACLog.Events.Disconnected += clientACLog_Disconnected;
			clientACLog.Events.DataReceived += clientACLog_DataReceived;
			clientACLog.Events.DataSent		+= clientACLog_DataSent;
			clientACLog.Keepalive.EnableTcpKeepAlives = true;
			clientACLog.Settings.MutuallyAuthenticate = false;
			clientACLog.Settings.AcceptInvalidCertificates = true;
			clientACLog.Settings.ConnectTimeoutMs = 5000;
			clientACLog.Logger = Logger;

			try {
				clientUDP = new UdpEndpoint("", Convert.ToInt32(Settings1.Default.URCPort) + 1);
				clientUDP.EndpointDetected += clientUDP_EndpointDetected;

				// only if you want to receive messages...
				clientUDP.DatagramReceived += clientUDP_DatagramReceived;
				clientUDP.Start();
			}
			catch (Exception ex){ 
			}

			this.TopMost = true;


		}

		private void Form1_Load(object sender, EventArgs e) {

			this.Location = Settings1.Default.MainLocation;
			// Sometimes we have seen a wrong size (very small) during debug, this takes case of this
			if (Settings1.Default.MainSize.Height >= this.Size.Height && Settings1.Default.MainSize.Width >= this.Size.Width)
				this.Size = Settings1.Default.MainSize;

			tbURCIP.Text = Settings1.Default.URCIP;
			tbURCPort.Text = Settings1.Default.URCPort;
			tbACLogPort.Text = Settings1.Default.ACLogPort;
			initDone = true;
		}


		private void Form1_LocationChanged(object sender, EventArgs e) {
			if (initDone)
				saveWindow();

		}

		private void Form1_SizeChanged(object sender, EventArgs e) {
			if (initDone)
				saveWindow();

		}
		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {
			Settings1.Default.ACLogPort = tbACLogPort.Text;

			saveWindow();
		}

		//-------------------------------------------------------------------------
		// 
		//-------------------------------------------------------------------------
		private void saveWindow() {
			if (this.WindowState == FormWindowState.Normal) {
				// save location and size if the state is normal
				Settings1.Default.MainLocation = this.Location;
				Settings1.Default.MainSize = this.Size;

			} else {
				// save the RestoreBounds if the form is minimized or maximized!
				Settings1.Default.MainLocation = this.RestoreBounds.Location;
				Settings1.Default.MainSize = this.RestoreBounds.Size;
			}

			// don't forget to save the settings
			Settings1.Default.Save();

		}



		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		// SatPC32 stuff
		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		private void clientSatPC32_Advise(object sender, DdeAdviseEventArgs args) {

			try {
				if (ValidateIPv4(tbURCIP.Text)) {
					clientUDP.Send(Settings1.Default.URCIP, Convert.ToInt32(Settings1.Default.URCPort), args.Text);
				}
			}
			catch (Exception ex) { }

			Console.WriteLine("Data");


			// ** NO SATELLITE **\n\r\n
			if (args.Text.Contains("** NO SATELLITE **")) {
				// Clear fields
				setFieldACLog("PROPMODE", "");
				setFieldACLog("SATNAME", "");
				setFieldACLog("FREQUENCY", "");
				setFieldACLog("BAND", "");
				setFieldACLog("MODE", "");

				setFieldACLog(FREQ_RX, "");
				setFieldACLog(BAND_RX, "");

				lblSatName.Text = "NO SAT";
				lblDownlinkFreq.Text = "";
				lblDownlinkMode.Text = "";
				lblUplinkFreq.Text = "";
				lblUplinkMode.Text = "";
				lblAzimuth.Text = "";
				lblElevation.Text = "";

			} else {
				// Set "SAT" in PROPMODE field
				setFieldACLog("PROPMODE", "SAT");

				// SNAO-73 AZ91,7 EL1,9 UP435132865 UMLSB DN145962393 DMUSB MA209,5 RR-4,9154719\r\n
				// This will split to;´:
				// [0]	"SNQO-100"
				// [1] "AZ160,3"
				// [2] "EL24,9"
				// [3] "UP0"
				// [4] "UMLSB"
				// [5] "DN0"
				// [6] "DMUSB"
				// [7] "MA2,4"
				// [8] "RR-0,0002315\r\n

				String[] lines = args.Text.Split(" ");

				float freqUplink = Convert.ToInt64(lines[3].Substring(2), 10);
				freqUplink /= 1000000;
				float freqDownlink = Convert.ToInt64(lines[5].Substring(2), 10);
				freqDownlink /= 1000000;


				String satName = lines[0].Substring(2);
				// A little hack, in SatPC32 "ISS" is shown like that, also in the TLE files.
				// LoTW does not know "ISS", it needs "ARISS" instead, we fix that here :)
				if (satName == "ISS")
					satName = "ARISS";

				lblSatName.Text = lines[0].Substring(2);

				if (freqDownlink > 0) {
					lblDownlinkFreq.Text = freqDownlink.ToString("F6");
					lblDownlinkMode.Text = lines[6].Substring(2);
				} else {
					lblDownlinkFreq.Text = "";
					lblDownlinkMode.Text = "";
				}

				if (freqUplink > 0) {
					lblUplinkFreq.Text = freqUplink.ToString("F6");
					lblUplinkMode.Text = lines[4].Substring(2);
				} else {
					lblUplinkFreq.Text = "";
					lblUplinkMode.Text = "";
				}
			
				lblAzimuth.Text = lines[1].Substring(2);
				lblElevation.Text = lines[2].Substring(2);


				// Set SAT_NAME field
				setFieldACLog("SATNAME", satName);

				// Set FREQ, BAND and MODE fields
				if (freqUplink == 0.0) {
					setFieldACLog("FREQUENCY", "");
					setFieldACLog("BAND", "");
					setFieldACLog("MODE", "");
				} else {
					// Set the frequency in MHz
					setFieldACLog("FREQUENCY", freqUplink.ToString("F6"));
					// Set the correct BAND
					if (freqUplink >= 28 && freqUplink <= 30)
						setFieldACLog("BAND", "10");
					else if (freqUplink >= 144 && freqUplink <= 146)
						setFieldACLog("BAND", "2");
					else if (freqUplink >= 430 && freqUplink <= 440)
						setFieldACLog("BAND", "70");
					else if (freqUplink >= 1290 && freqUplink <= 1300)
						setFieldACLog("BAND", "23");

					if (lines[4] == "UMLSB" || lines[4]=="UMUSB")
						setFieldACLog("MODE", "SSB");
					else if (lines[4] == "UMFM")
						setFieldACLog("MODE", "FM");
					else if (lines[4] == "UMCW")
						setFieldACLog("MODE", "CW");
				}

				// Set FREQ_RX and BAND_RX fields
				if (freqUplink == 0.0) {
					setFieldACLog(BAND_RX, "");
					setFieldACLog(FREQ_RX, "");
				} else {

					// We replace "comma" with "decimal point" (depending on language settings in windows etc)
					// (could have been done more correct, just a quick hack ;)
					// ACLog/LoTW wants numbers with decimal point and not comma ;)
					String value = freqDownlink.ToString("F6");
					value = value.Replace(',', '.');

					// Set the receive frequency in MHz
					setFieldACLog(FREQ_RX, value);

					// Set the correct receive band
					if (freqDownlink >= 28 && freqDownlink <= 30)
						setFieldACLog(BAND_RX, "10M");
					else if (freqDownlink >= 144 && freqDownlink <= 146)
						setFieldACLog(BAND_RX, "2M");
					else if (freqDownlink >= 430 && freqDownlink <= 440)
						setFieldACLog(BAND_RX, "70CM");
					else if (freqDownlink >= 1290 && freqDownlink <= 1300)
						setFieldACLog(BAND_RX, "23CM");
				}
			}
		}

		private void clientSatPC32_Disconnected(object sender, DdeDisconnectedEventArgs args) {
			lblSatPC32Status.Text = "Disconnected";
		}



		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		// ACLog stuff
		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------

		//-----------------------------------------------------------------------------------------------
		// Cconnected to ACLog
		//-----------------------------------------------------------------------------------------------
		void clientACLog_Connected(object sender, ConnectionEventArgs e) {
			// As we are being called from another thread we need to Invoke the methods..
			this.Invoke(new MethodInvoker(delegate {
				//clientACLog.Send("<CMD><PROGRAM></CMD>" + "\r\n");  //Must include \r\n for the client to send the command!
				lblACLogStatus.Text = "Connected";
			}));
		}

		//-----------------------------------------------------------------------------------------------
		// Disconnected from ACLog
		//-----------------------------------------------------------------------------------------------
		void clientACLog_Disconnected(object sender, ConnectionEventArgs e) {
			// As we are being called from another thread we need to Invoke the methods..
			this.Invoke(new MethodInvoker(delegate {
				//		tbACLog.Text = "*** Server " + e.IpPort + " disconnected";
				lblACLogStatus.Text = "Disconnected";
			}));
		}

		//-----------------------------------------------------------------------------------------------
		// Data received from ACLog
		//-----------------------------------------------------------------------------------------------
		void clientACLog_DataReceived(object sender, DataReceivedEventArgs e) {
			// As we are being called from another thread we need to Invoke the methods..
			this.Invoke(new MethodInvoker(delegate {
		//		tbACLogResp.Text = "[" + e.IpPort + "] " + Encoding.UTF8.GetString(e.Data.Array, 0, e.Data.Count);

			}));
		}

		//-----------------------------------------------------------------------------------------------
		// Called when data has been sent to ACLog
		//-----------------------------------------------------------------------------------------------
		private void clientACLog_DataSent(object sender, DataSentEventArgs e) {
			// As we are being called from another thread we need to Invoke the methods..
			this.Invoke(new MethodInvoker(delegate {
		//		tbACLog.Text = "[" + e.IpPort + "] sent " + e.BytesSent + " bytes";
			}));
		}

		 void Logger(string msg) {
			Console.WriteLine(msg);
		}


		//-----------------------------------------------------------------------------------------------
		// Set a field in ACLog to a specific value
		//-----------------------------------------------------------------------------------------------
		private void setFieldACLog(String field, String value) {
			if (lblACLogStatus.Enabled == false)
				return;
			if (clientACLog.IsConnected) clientACLog.Send("<CMD><UPDATE><CONTROL>TXTENTRY" + field +"</CONTROL><VALUE>" + value + "</VALUE></CMD>" + "\r\n");
		}


		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		// UDP Client handlers
		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		static void clientUDP_EndpointDetected(object sender, EndpointMetadata md) {
			Console.WriteLine("Endpoint detected: " + md.Ip + ":" + md.Port);
		}

		static void clientUDP_DatagramReceived(object sender, Datagram dg) {
			Console.WriteLine("[" + dg.Ip + ":" + dg.Port + "]: " + Encoding.UTF8.GetString(dg.Data));
		}


		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		// Timer handlers
		//-----------------------------------------------------------------------------------------------
		//-----------------------------------------------------------------------------------------------
		private void timerSatPC32_Tick(object sender, EventArgs e) {
			if (clientSatPC32.IsConnected == false) {
				try {
					// Connect to the server.  It must be running or an exception will be thrown.
					clientSatPC32.Connect();

					// Advise Loop
					//client.StartAdvise("myitem", 1, true, 60000);
					clientSatPC32.StartAdvise("SatPcDdeItem", 1, true, 60000);
					lblSatPC32Status.Text = "Connected";

				}
				catch (Exception ex) {
					lblSatPC32Status.Text = "Disconnected";
				}
			}
		}

		//-----------------------------------------------------------------------------------------------
		// Check if a string is a valid IP4 address
		//-----------------------------------------------------------------------------------------------
		public bool ValidateIPv4(string ipString) {
			if (String.IsNullOrWhiteSpace(ipString)) {
				return false;
			}

			string[] splitValues = ipString.Split('.');
			if (splitValues.Length != 4) {
				return false;
			}

			byte tempForParsing;

			return splitValues.All(r => byte.TryParse(r, out tempForParsing));
		}

		private void tbURCIP_KeyUp(object sender, KeyEventArgs e) {

			if (ValidateIPv4(tbURCIP.Text)) {
				Settings1.Default.URCIP = tbURCIP.Text;
				Settings1.Default.URCPort = tbURCPort.Text;
				Settings1.Default.Save();
			}

		}

		private void btnChangeURC_Click(object sender, EventArgs e) {
			if (tbURCIP.Enabled) {
				if (ValidateIPv4(tbURCIP.Text)) {
					Settings1.Default.URCIP = tbURCIP.Text;
					Settings1.Default.URCPort = tbURCPort.Text;
					Settings1.Default.Save();
					tbURCIP.Enabled = false;
					tbURCPort.Enabled = false;
				}
			} else {
				tbURCIP.Enabled = true;
				tbURCPort.Enabled = true;
			}

		}

		private void timerACLog_Tick(object sender, EventArgs e) {
			if (clientACLog.IsConnected == false) {
				try {
					clientACLog.ConnectWithRetries(500);
				}
				catch (Exception ex) {

				}
			}

		}

		private void cbACLog_CheckedChanged(object sender, EventArgs e) {

			if (cbACLog.Checked == false) {
				// Clear fields
				setFieldACLog("PROPMODE", "");
				setFieldACLog("SATNAME", "");
				setFieldACLog("FREQUENCY", "");
				setFieldACLog("BAND", "");
				setFieldACLog("MODE", "");

				setFieldACLog(FREQ_RX, "");
				setFieldACLog(BAND_RX, "");
			}

			lblACLogStatus.Enabled = cbACLog.Checked;



		}
	}
}
